#include "regalloc.hpp"

#include "Function.h"
#include "Instruction.h"
#include "liverange.hpp"

#include <algorithm>

using std::for_each;

using namespace RA;

int
get_arg_id(Argument *arg) {
    auto args = arg->get_parent()->get_args();
    int id = 1;
    for (auto a : args) {
        if (a == arg)
            break;
        ++id;
    }
    return id;
}

bool
RegAllocator::no_reg_alloca(Value *v) const {
    auto instr = dynamic_cast<Instruction *>(v);
    auto arg = dynamic_cast<Argument *>(v);
    if (instr)
        return instr->is_alloca() or instr->is_cmp() or instr->is_fcmp() or
               instr->is_zext();
    if (arg) {
        return get_arg_id(arg) > ARG_MAX_R;
    } else
        assert(false && "only instruction and argument's LiveInterval exits");
}

void
RegAllocator::reset(Function *func) {
    cur_func = func;
    regmap.clear();
    active.clear();
    for_each(used, used + R + 1, [](bool &u) { u = false; });
}

int
RegAllocator::ReserveForArg(const LVITS &Liveints) {
    auto args = cur_func->get_args();
    auto it_int = Liveints.begin();
    auto it_arg = args.begin();
    int reg;
    for (reg = 1; reg <= args.size() and reg <= ARG_MAX_R; ++reg) {
        auto arg = *it_arg;
        auto liveint = *it_int;
        assert(arg == liveint.second && "arg should be in order in liveints");

        used[reg] = true;
        regmap[arg] = reg;
        active.insert(liveint);
        ++it_arg, ++it_int;
    }
    return reg;
}

void
RegAllocator::LinearScan(const LVITS &liveints, Function *func) {
    reset(func);
    ReserveForArg(liveints);
    int reg;
    for (auto liveint : liveints) {
        if (dynamic_cast<Argument *>(liveint.second)) {
            continue;
        }
        if (no_reg_alloca(liveint.second))
            continue;
        ExpireOldIntervals(liveint);
        if (active.size() == R)
            SpillAtInterval(liveint);
        else {
            for (reg = 1; reg <= R and used[reg]; ++reg)
                ;
            used[reg] = true;
            regmap[liveint.second] = reg;
            active.insert(liveint);
        }
    }
}

void
RegAllocator::ExpireOldIntervals(LiveInterval liveint) {
    auto it = active.begin();
    for (; it != active.end() and it->first.j < liveint.first.i; ++it)
        used[regmap.at(it->second)] = false;
    active.erase(active.begin(), it);
}

void
RegAllocator::SpillAtInterval(LiveInterval liveint) {
    auto spill = *active.rbegin();
    if (dynamic_cast<Argument *>(spill.second))
        return;
    if (spill.first.j > liveint.first.j) {
        // cancel reg allocation for spill
        regmap[liveint.second] = regmap.at(spill.second);

        active.erase(spill);
        regmap.erase(spill.second);
    }
}
